/**
 * Copyright (C) 2007-2008 by Seiko Instruments Inc.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation under the terms of the GNU General Public License is hereby
 * granted. No representations are made about the suitability of this software
 * for any purpose. It is provided "as is" without express or implied warranty.
 * See the GNU General Public License for more details.
 */

/// Model
#define CAPD245					8245
#define CAPD247					8247
#define CAPD345					8345
#define CAPD347					8347

/// MaxWidth
#define DEFAULTMAXWIDTH			54*8
#define CAPD245MAXWIDTH			48*8
#define CAPD247MAXWIDTH			54*8
#define CAPD345MAXWIDTH			72*8
#define CAPD347MAXWIDTH			72*8

/// Setting Parameter
#define PAGECUT					0
#define DOCCUT					1
#define BLANKFEED				0
#define BLANKNONFEED			1
#define FULLCUT					0
#define PARTIALCUT				1
#define NOCUT					2
#define NORMAL					0
#define DRAFT					3
#define DITHERON				0
#define DITHEROFF				1
#define NORMAL_MODE_IMAGE		0
#define DRAFT_MODE_IMAGE		3
#define REDUCEDOFF				0
#define REDUCEDON				1
#define THRESHOLDVAL			0x7F
#define BINMAXVAL				0xFF
#define BINMINVAL				0x00

/// Reduce Page Size
static const char *pszReducePageSize[] =
{
	"A4",
	"Letter",
	NULL
};

/// Printer Base pitch
#define PRTPITCHNORMAL			203
#define PRTPITCHDRAFT			101

/// Printer Seppd
#define PRTSPEED_NUM			2
#define PRTSPEEDLOW				1
#define PRTSPEEDHIGH			0

// Printer Density
#define PRTDENSITY_NUM			9
#define PRTDENSITY60			0
#define PRTDENSITY70			1
#define PRTDENSITY80			2
#define PRTDENSITY90			3
#define PRTDENSITY100			4
#define PRTDENSITY110			5
#define PRTDENSITY120			6
#define PRTDENSITY130			7
#define PRTDENSITY140			8


/// Printer Cut mode
#define PRTCUTFULL				0
#define PRTCUTPARTIAL			1

#define SENDBEFOREFULLCUT 		1
#define SENDAFTERFULLCUT		2
#define SENDBEFOREPARTIALCUT	3
#define SENDAFTERPARTIALCUT		4
#define SENDBEFORENONCUT		5
#define SENDAFTERNONCUT			6


#define OUTPUT_LINE_NUM			20


typedef struct tagSendLength_t
{
	int nModelNumber;
	int nBeforeFullCut;
	int nAfterFullCut;
	int nBeforePartialCut;
	int nAfterPartialCut;
	int nBeforeNonCut;
	int nAfterNonCut;
} SENDLENGTH;

static const SENDLENGTH tblSendLength[] =
{
	{CAPD245, 55, 12, 55, 12,  0,  67},
	{CAPD247, 55, 12, 55, 12,  0,  67},
	{CAPD345, 55, 12, 55, 12,  0,  67},
	{CAPD347, 55, 12, 55, 12,  0,  67},
	{0, 0, 0, 0, 0, 0, 0}
};


typedef struct tagMaxBytesPerLine_t
{
	int nModelNumber;
	int nMaxBytesPerLine;
} MAXBYTELENGTH;

static const MAXBYTELENGTH tblMaxBytesPerLine[] =
{
	{CAPD245, CAPD245MAXWIDTH},
	{CAPD247, CAPD247MAXWIDTH},
	{CAPD345, CAPD345MAXWIDTH},
	{CAPD347, CAPD347MAXWIDTH},
	{0, 0}
};

static const unsigned char data_shift_tbl[] = {
	0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01
};

typedef struct tagPrintConditions_t
{
	int nModelNumber;
	int nSpeed[ PRTSPEED_NUM ];
	int nPrintDensity[ PRTDENSITY_NUM ];
} PRINTCONDITIONS;


static const PRINTCONDITIONS tblPrintConditions[] =
{
	{CAPD245, {0, 0}, {60,70,80,90,100,110,120,130,140} },
	{CAPD247, {0, 0}, {60,70,80,90,100,110,120,130,140} },
	{CAPD345, {0, 0}, {60,70,80,90,100,110,120,130,140} },
	{CAPD347, {0, 0}, {60,70,80,90,100,110,120,130,140} },
	{0, {0, 0}, {0,0,0,0,0,0,0,0,0} }
};
